# Clamix
# clustering symbolic objects described by discrete distributions
# V. Batagelj, S. Korenjak-erne, N. Kejar
# University of Ljubljana
# july 2010
# october 2012: specificity
# -------------------------------------------------------------------
# parameters
#   nVar  - number of variables in SO description
#   alpha - weights in dissimilarity
#   so    - empty SO
# are provided as global variables

# http://www.math.tau.ac.il/~yekutiel/CDA/Thompson manual.pdf
# page 10 & 11
# http://books.google.com/books?id=hpEzw4T0sPUC&printsec=frontcover&dq=analysis+of+categorical+data
# page 20
# V. Batagelj / 28.7.2010
# ---------------------------------------------------------------------
acceptbin<-function(x,n,p){
# computes the acceptability of p when x is observed and X is Bin(n,p)
# adapted from Blaker (2000)
  p1 <- 1-pbinom(x-1,n,p)
  p2 <- pbinom(x,n,p)
  a1 <- p1 + pbinom(qbinom(p1,n,p)-1,n,p)
  a2 <- p2 + 1 - pbinom(qbinom(1-p2,n,p),n,p)
  min(a1,a2)
}
acceptinterval<-function(x, n, level, tolerance=1e-04){
# computes acceptability interval for p at 1-alpha equal to level
# (in (0,1)) when x is an observed value of X which is Bin(n,p)
# adapted from Blaker (2000)
  lower <- 0; upper <- 1
  if(x) {lower <- qbeta((1-level)/2,x,n-x+1)
    while(acceptbin(x,n,lower) < (1-level)) lower <- lower+tolerance
  }
  if(x!=n) {upper <- qbeta(1-(1-level)/2,x+1,n-x)
    while(acceptbin(x,n,upper) < (1-level)) upper <- upper-tolerance
  }
  c(lower=lower,upper=upper)
}
# acceptinterval(x=0,n=25,level=.95)
# lower upper
# 0 0.1275852

# -----------------------------------------------------------------------

# creates an empty SO
emptySO <- function(nCats){
  nVar <- length(nCats)
  s <- vector("list", nVar)
  for(i in 1:nVar) s[[i]] <- double(nCats[i]+1)
  return(s)
}

# computes weighted squared Euclidean dissimilarity between SOs
# global: nVar, alpha
distSO <- function(X,Y){
  d <- numeric(nVar)
  for(i in 1:nVar) {
    ln <- length(X[[i]]);
    nX <- as.double(X[[i]][[ln]]); nY <- as.double(Y[[i]][[ln]])
    d[[i]] <- nX*sum((X[[i]][-ln]/nX-Y[[i]][-ln]/nY)**2)
  }
  dis <- as.numeric(d %*% alpha)/2
  if (is.na(dis)) dis <- Inf
  return(dis)
}

# computes dissimilarity between clusters
# global: nVar, alpha
distCl <- function(X,Y){
  d <- numeric(nVar)
  for(i in 1:nVar) {
    ln <- length(X[[i]]);
    nX <- as.double(X[[i]][[ln]]); nY <- as.double(Y[[i]][[ln]])
    d[[i]] <- nX*nY/(nX+nY)*sum((X[[i]][-ln]/nX-Y[[i]][-ln]/nY)**2)
  }
  return(as.numeric(d %*% alpha)/2)
}

# encode numerical vector using given encoding
encodeSO <- function(x,encoding,codeNA){
  if(is.na(x)) return(codeNA)
  for(i in 1:length(encoding)) if(encoding[[i]](x)) return(i)
}

# adapted leaders method
#---------------------------------------------
# VB, 16. julij 2010
#   dodaj omejitev na najmanje tevilo enot v skupini
#   omejeni polmer
# global: so, numSO, nVar, alpha
leaderSO <- function(SOs,maxL){
  numSO <- length(SOs)
  L <- vector("list",maxL); Ro <- numeric(maxL)
# random partition into maxL clusters
  clust <- sample(1:maxL,numSO,replace=TRUE)
  tim <- 1; step <- 0
  repeat {
    step <- step+1
  # new leaders - determine the leaders of clusters in current partition
    for(k in 1:maxL){L[[k]] <- so; names(L)[[k]] <- paste("L",k,sep="")}
    for(i in 1:numSO){j <- clust[[i]];
      for(k in 1:nVar) L[[j]][[k]] <- L[[j]][[k]] + SOs[[i]][[k]] }
  # new partition - assign each unit to the nearest new leader
  # for(k in 1:maxL) print(L[[k]]); flush.console() # TEST
    clust <- integer(numSO)
    R <- numeric(maxL); p <- double(maxL)
    for(i in 1:numSO){d <- double(maxL)
      for(k in 1:maxL){d[[k]] <- distSO(SOs[[i]],L[[k]])}
      r <- min(d); j <- which(d==r)
      if(is.infinite(r)){cat("Infinite unit=",i,"\n"); print(SOs[[i]]); flush.console()}
      if(length(j)==0){
        cat("unit",i,"\n",d,"\n"); flush.console(); print(SOs[[i]]); flush.console()
        u <- which(is.na(d))[[1]]; cat("leader",u,"\n"); print(L[[u]])
        stop()}
      j <- which(d==r)[[1]];
      clust[[i]] <- j
      p[[j]] <- p[[j]] + r; if(R[[j]]<r) R[[j]]<- r
    }
  # report
    cat("\nStep",step,"\n"); flush.console()
    print(table(clust)); print(R); print(Ro-R); Ro <- R;
    print(p); print(sum(p)); flush.console()
    tim <- tim-1
    if(tim<1){
      ans <- readline("Times repeat = ")
      tim <- as.integer(ans); if (tim < 1)  break
    }
  # in the case of empty cluster put in the most distant SO
    repeat{
      t <- table(clust); em <- setdiff(1:maxL,as.integer(names(t)))
      if(length(em)==0) break
      j <- em[[1]]; rmax <- 0; imax <- 0
      for(i in 1:numSO){d <- double(maxL)
        for(k in 1:maxL){d[[k]] <- distSO(SOs[[i]],L[[k]])}
        r <- max(d);  if(rmax<r) {rmax <- r; imax <- i}
      }
      clust[[imax]] <- j; L[[j]] <- SOs[[imax]]
      cat("*** empty cluster",j,"- SO",imax,"transfered, rmax =",rmax,"\n")
      flush.console()
    }
  }
  return (list(clust=clust,leaders=L,R=R,p=p))
}

# adapted Ward's hierarchical clustering
#---------------------------------------------
# VB, 16. julij 2010
#   imena notranjih tok v drevesu
# global: so, nVar, alpha
hclustSO <- function(L){
  orDendro <- function(i){if(i<0) return(-i)
    return(c(orDendro(m[i,1]),orDendro(m[i,2])))}
  numL <- length(L); numLm <- numL-1
# each unit is a cluster; compute inter-cluster dissimilarity matrix
  D <- matrix(nrow=numL,ncol=numL); diag(D) <- Inf
  for(i in 1:numLm) for(j in (i+1):numL) {
      D[i,j] <- distCl(L[[i]],L[[j]]); D[j,i] <- D[i,j]
  }
  active <- 1:numL; m <- matrix(nrow=numLm,ncol=2)
  node <- rep(0,numL); h <- numeric(numLm); LL <- vector("list",numLm)
  for(k in 1:numLm) LL[[k]] <- so
  for(k in 1:numLm){
  # determine the closest pair of clusters (p,q)
    ind <- active[sapply(active,function(i) which.min(D[i,active]))]
    dd <- sapply(active,function(i) min(D[i,active]))
    pq <- which.min(dd)
  # join the closest pair of clusters
    p<-active[pq]; q <- ind[pq]; h[k] <- D[p,q]
    if(node[p]==0){m[k,1] <- -p; Lp <- L[[p]]
      } else {m[k,1] <- node[p]; Lp <- LL[[node[p]]]}
    if(node[q]==0){m[k,2] <- -q; Lq <- L[[q]]
      } else {m[k,2] <- node[q]; Lq <- LL[[node[q]]]}
    for(t in 1:nVar) LL[[k]][[t]] <- Lp[[t]] + Lq[[t]]
    active <- setdiff(active,p)
    Lpq <- LL[[k]]
  # determine dissimilarities to the new cluster
    for(s in setdiff(active,q)){
      if(node[s]==0){Ls <- L[[s]]} else {Ls <- LL[[node[s]]]}
      D[q,s] <- distCl(Lpq,Ls); D[s,q] <- D[q,s]
    }
    node[[q]] <- k
  }
  hc <- list(merge=m,height=h,order=orDendro(numLm),labels=names(L),
    method=NULL,call=NULL,dist.method=NULL,leaders=LL)
  class(hc) <- "hclust"
  return(hc)
}

# prints values that are significantely different from the average
testSOvar <- function(SO,var,total,a){
  ln <- length(total[[var]]); lnM <- ln-1
  pt <- total[[var]][-ln]/total[[var]][[ln]]
  pj <- SO[[var]][-ln]/SO[[var]][[ln]]
  for(k in 1:lnM) {
    ai <- acceptinterval(x=total[[var]][[k]], n=total[[var]][[ln]], level=1-a)
    if(pj[[k]] > ai[[2]]) cat(format(names(total[[var]])[[k]],width=15,justify="right"),
      format(c(pj[[k]],ai[[2]],pt[[k]],ai[[1]],(pj[[k]]-pt[[k]])/(ai[[2]]-pt[[k]])),
      width=12,justify="left",digits=5,nsmall=7),"\n")
    if(pj[[k]] < ai[[1]]) cat(format(names(total[[var]])[[k]],width=15,justify="right"),
      format(c(pj[[k]],ai[[2]],pt[[k]],ai[[1]],(pt[[k]]-pj[[k]])/(pt[[k]]-ai[[1]])),
      width=12,justify="left",digits=5,nsmall=7),"\n")
  }
}

# prints values that are significantely larger than the average
testSOvarP <- function(SO,var,total,a){
  ln <- length(total[[var]]); lnM <- ln-1
  pt <- total[[var]][-ln]/total[[var]][[ln]]
  pj <- SO[[var]][-ln]/SO[[var]][[ln]]
  for(k in 1:lnM) {
    ai <- acceptinterval(x=total[[var]][[k]], n=total[[var]][[ln]], level=1-a)
    if(pj[[k]] > ai[[2]]) cat(format(names(total[[var]])[[k]],width=15,justify="right"),
      format(c(pj[[k]],ai[[2]],pt[[k]],ai[[1]],(pj[[k]]-pt[[k]])/(ai[[2]]-pt[[k]])),
      width=12,justify="left",digits=5,nsmall=7),"\n")
  }
}

# prints values that are significantely different from the average
testSOvarQ <- function(SO,var,total,a,qmin){
  ln <- length(total[[var]]); lnM <- ln-1
  pt <- total[[var]][-ln]/total[[var]][[ln]]
  pj <- SO[[var]][-ln]/SO[[var]][[ln]]
  for(k in 1:lnM) {
    ai <- acceptinterval(x=total[[var]][[k]], n=total[[var]][[ln]], level=1-a)
    if(pj[[k]] > ai[[2]]) {q <- (pj[[k]]-pt[[k]])/(ai[[2]]-pt[[k]])
      if(q>qmin) cat(format(names(total[[var]])[[k]],width=15,justify="right"),
      format(c(pj[[k]],ai[[2]],pt[[k]],ai[[1]],q),
      width=12,justify="left",digits=5,nsmall=7),"\n")}
    if(pj[[k]] < ai[[1]]) {q <- (pt[[k]]-pj[[k]])/(pt[[k]]-ai[[1]])
      if(q>qmin) cat(format(names(total[[var]])[[k]],width=15,justify="right"),
      format(c(pj[[k]],ai[[2]],pt[[k]],ai[[1]],q),
      width=12,justify="left",digits=5,nsmall=7),"\n")}
  }
}

# prints "equiprobabilistic" encodings for numerical variables
makeEnc <- function(var,name,k){
  n <- length(var)-length(which(is.na(var)))
  v <- sort(var)
  steps <- seq.int(0,n,n %/% (k-1))
  cuts <- c(0,v[steps[1:(k-1)]],max(v,na.rm=TRUE))
  cat("enc",name," <- list(\n",sep="")
  cat('  "[0]" = function(x) x<=0,\n')
  for(j in 2:k) cat('  "(',cuts[j-1],',',cuts[j],']" = function(x) x<=',cuts[j],',\n',sep='')
  cat('  "NA" = function(x) TRUE )\n')
}

# computes totals
computeTotal <- function(L){
  numL <- length(L); total <- namedSO
  for(i in 1:numL) for(j in 1:nVar) total[[j]] <- total[[j]] + L[[i]][[j]]
  return(total)
}

plotDistri <- function(pU,pC){
  ln <- length(pU)
  plot(c(1,ln),c(0,max(max(pU),max(pC))),type="n",main="Distributions",
    xlab="values",ylab="p")
  lines(c(0,pU[-ln]),type="S",col="blue")
  lines(c(ln,ln),c(pU[ln-1],0),col="blue")
  lines((1:ln)+0.02,c(0,pC[-ln]),type="S",col="red")
  lines(c(ln,ln)+0.02,c(pC[ln-1],0),col="red")
}

specific <- function(leader,var){
  Lq <- L[[leader]]
  names(Lq) <- names(total)
  q <- Lq[[var]]
  names(q) <- names(pU[[var]])
  ln <- length(q)
  q <- q/q[ln]
  c <- q[-ln]
  u <- pU[[var]][-ln]
  print(u)
  print(c)
  print(pmax(u,c)/pmin(u,c))
  plotDistri(pU[[var]],q)
}
